----------------------------------------------------------------------------------------------------
-- Pixel Counter for 256 x 192 Pixel Display on TFT 480x234
-- by C.Meyer & c't magazin 4/2009
-- FRAME_ON should be high during 256 clock cycles horizontal, 192 lines vertical
----------------------------------------------------------------------------------------------------
library ieee;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use ieee.numeric_std.all;

entity framer is
  port (
    PCLK: in  std_logic;
	HPIXEL: in  std_logic_vector(9 downto 0); -- max. 1024 x 1024 Pixel
	VPIXEL: in  std_logic_vector(9 downto 0); -- these are valid 2 PCLKs before pixel output
	GRIDX: in std_logic_vector (7 downto 0);
	GRIDY: in std_logic_vector (7 downto 0);
	SMALLW: in  std_logic;	-- not used here
	GSTARTX: 	in std_logic_vector (7 downto 0);	-- graph window borders
	GSTARTY:	in std_logic_vector (7 downto 0);
	GSIZEX:	in std_logic_vector (7 downto 0);
	GSIZEY:	in std_logic_vector (7 downto 0);
    GRIG_ON:	out  std_logic;
	FRAME_ON:	out std_logic;
	PLOTSYNC:	out std_logic;	-- Sync fr Plot-Routinen in CPU-Core
    GADDR: 	out std_logic_vector (15 downto 0)
    );
end framer;

architecture behave of framer is

constant last_line: STD_LOGIC_VECTOR(9 DOWNTO 0) := CONV_STD_LOGIC_VECTOR(233, 10);	-- for screen height 234 pixel
signal frame_x : std_logic_vector (9 downto 0);
signal frame_y : std_logic_vector (9 downto 0);
signal hgridcount  : integer range 255 downto 0;
signal vgridcount  : integer range 255 downto 0;
signal clk, pclk_half : std_logic;
signal xvalid : std_logic;
signal yvalid : std_logic;
signal framevalid : std_logic;
signal vborder : std_logic;
signal hborder : std_logic;
signal hgrid : std_logic;
signal vgrid : std_logic;
signal hwinstart  : std_logic_vector (9 downto 0);
signal vwinstart  : std_logic_vector (9 downto 0);
signal hwinend, hwinend_border : std_logic_vector (9 downto 0);
signal vwinend : std_logic_vector (9 downto 0);
  
begin
clk <= PCLK; -- when (SMALLW='1') else pclk_half;
--Clock divider /2
--pclk_half <= (not PCLK) when rising_edge(PCLK);
hwinstart <= ('0' & GSTARTX & '0');
vwinstart <= ('0' & GSTARTY & '0');
hwinend <= ('0' & GSTARTX & '0')+('0' & GSIZEX & '0');
vwinend <= ('0' & GSTARTY & '0')+('0' & GSIZEY & '0');
hwinend_border <= hwinend-1;
frame_x <= HPIXEL - hwinstart;
frame_y <= VPIXEL - vwinstart;

GADDR(7 downto 0) <= frame_x(7 downto 0);
GADDR(15 downto 8) <= frame_y(7 downto 0);

framevalid	<=	xvalid and yvalid;
FRAME_ON<=framevalid;
GRIG_ON <= (vborder or hborder or hgrid or vgrid) and framevalid;

process (clk)	-- horizontale X-Position innerhalb FRAME_ON hochzhlen
begin
	if xvalid = '0' then
		hgridcount<=1;
		hgrid<='0';
	elsif rising_edge(clk) then 
		hgridcount <= hgridcount+1;
		if hgridcount=GRIDX then
			hgridcount<=1;
			hgrid<='1';
		else
			hgrid<='0';
		end if;
	end if;
end process;

process (xvalid, yvalid)	-- horizontale X-Position innerhalb FRAME_ON hochzhlen
begin
	if yvalid = '0' then
		vgridcount<=1;
		vgrid<='0';
	elsif falling_edge(xvalid) then 
		vgridcount <= vgridcount+1;
		if vgridcount=GRIDY then
			vgridcount<=1;
			vgrid<='1';
		else
			vgrid<='0';
		end if;
	end if;
end process;

valids: process (PCLK)
begin
	if rising_edge(PCLK) then
		if VPIXEL=last_line then
			PLOTSYNC <=	'1';
		else
			PLOTSYNC <=	'0';
		end if;
		hborder<='0';
		vborder<='0';
		if VPIXEL=vwinstart then
			yvalid<='1';
			vborder<='1';
		elsif VPIXEL=vwinend then
			vborder<='1';
		elsif VPIXEL>vwinend then
			yvalid<='0';
		end if;
		if HPIXEL=hwinstart then
			hborder<='1';
			xvalid<='1';
		elsif HPIXEL=hwinend_border then
			hborder<='1';
		elsif HPIXEL=hwinend then
			xvalid<='0';
		end if;
	end if;
end process;

end behave;
