----------------------------------------------------------------------------------
-- Company:        c't magazin
-- Engineer:       Carsten Meyer, cm@ctmagazin.de
-- 
-- Create Date:    14:11:15 06/05/2010 
-- Design Name: 
-- Module Name:    tft_text - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description:		Timing-Interface for analog LC-Display 7" DPF 480*234 or others
--					with externally supplied clock/sync-signals e.g. from digital picture frame
--					Display Pixel Data valid on falling PCLK edge
--					Selectable font size
-- Dependencies: 
--
-- Revision: 
-- Revision 0.90 - Font size adjustable by constants
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use ieee.numeric_std.all;
Library UNISIM;
use UNISIM.vcomponents.all;

entity tft_text is
    Port (	PCLK_i : in  STD_LOGIC;			-- pixel clock (about 10 MHz on DPF display 480 x 234), R75 on Haier Hi1018 PCB
			VSTART_i : in  STD_LOGIC; 		-- vertical start from display ca. 55 Hz, R71 on Haier PCB
			HSTART_i : in  STD_LOGIC;		-- horizontal start from display ca. 14,7 kHz, R74 on Haier PCB
			-- VCOM_i : in  STD_LOGIC:='0';	-- AC Component Vcom from display to invert RGB video, EC7 low side on Haier PCB
			
			RST    : in  std_logic;
			SYSCLK50 : in  std_logic;	-- 50 MHz system clock
			TEXTRAM_WR : in  std_logic;
			RAMDASCII: in  std_logic_vector(7 downto 0);
			READASCII: out  std_logic_vector(7 downto 0);
			READCOLOR: out  std_logic_vector(3 downto 0);
			TEXTX: in  std_logic_vector(7 downto 0);
			TEXTY: in  std_logic_vector(7 downto 0);
			CURSTYLE: in  std_logic_vector(2 downto 0);
			COLOR: in  std_logic_vector(3 downto 0); 	-- Invert (3), R(2), G(1), B(0)
			
			HPIXEL: out  std_logic_vector(9 downto 0); -- max. 1024 x 1024 Pixel
			VPIXEL: out  std_logic_vector(9 downto 0);
			PIXELADR: out  std_logic_vector(19 downto 0); -- these are valid 2 PCLKs before pixel output
			
			FRAME	: out std_logic;	-- frame valid output, '1' on output of pixel '0'
			RED		: out STD_LOGIC;	-- color pixel output, to 3 transistor driver base C on Haier PCB
			GREEN	: out STD_LOGIC;
			BLUE	: out STD_LOGIC;
			PCLK	: out std_logic;	-- synchronized PCLK
			HTICK	: out std_logic;	-- both ticks delayed to real frame start
			VTICK	: out std_logic		-- sync'd with htick
		   );
end tft_text;


architecture Behavioral of tft_text is

-- ############################ ADJUST THESE CONSTANTS TO YOUR DISPLAY REQUIREMENTS ###############################
constant screen_width: STD_LOGIC_VECTOR(9 DOWNTO 0) := CONV_STD_LOGIC_VECTOR(480, 10); -- screen width 480 pixel
constant screen_height: STD_LOGIC_VECTOR(9 DOWNTO 0) := CONV_STD_LOGIC_VECTOR(234, 10);	-- screen height 234 pixel
constant ofs_v: integer:= 3;	-- vertical position
constant ofs_h: integer:= 78; 	-- horizontal position, may be > screen_height for "negative" delay, 78 for DPF

-- adust these to your font requirements
-- 8x12 font, 60 chars @ 480 hor pixels, 19 text lines
constant char_height: integer:= 12;	-- 12 for 8x12 font @ 60 chars
constant char_width: integer:= 8; 	-- 8 for 8x12 font
constant col_h: STD_LOGIC_VECTOR(6 DOWNTO 0) := CONV_STD_LOGIC_VECTOR(60, 7);	-- text width 60 chars

-- 6x8 font, 80 chars @ 480 hor pixels
--constant char_height: integer:= 12;	-- 12 for 8x12 font @ 60 chars
--constant char_width: integer:= 8; 	-- 8 for 8x12 font
--constant col_h: STD_LOGIC_VECTOR(6 DOWNTO 0) := CONV_STD_LOGIC_VECTOR(80, 7);	-- text width 80 chars

-- ###############################################################################################################

-- may need tweaking if larger displays than 1024x1024 used
signal pixel_column, pixel_row : STD_LOGIC_VECTOR(9 DOWNTO 0) :=(others =>'0'); -- Pixel-Position zwei PCLKs vorher
signal pixel_adr_intern : STD_LOGIC_VECTOR(19 DOWNTO 0) :=(others =>'0'); -- absolute pixel address
signal pclk_intern : STD_LOGIC;
signal vstart, vstart_preset, vstart_trigger, vstart_d, hstart,hstart_d : STD_LOGIC:='0';	-- Impulse 0,5 PCLK delayed
signal cursor_on : STD_LOGIC:='0';
-- signal pclk_n : STD_LOGIC;
-- signal clkfb, CLK2X180, CLKFX,CLKFX180: STD_LOGIC;
signal pixel, pixel_red, pixel_blue, pixel_green, h_valid, v_valid, frame_intern : STD_LOGIC:='0';

-- interne Signale zur Adressberechnung
-- signal textX_intern,textY_intern,btnX_intern,btnY_intern : std_logic_vector(7 downto 0);

begin

HTICK <= hstart;
VTICK <= vstart;
HPIXEL <= pixel_column;
VPIXEL <= pixel_row;
PCLK <= pclk_intern;
PIXELADR <= pixel_adr_intern;
FRAME <= frame_intern;

-- assign R(2), G(1), B(0) from color-MEM
RED <= 		pixel_red;	-- pixel_red xnor VCOM_i; -- test
GREEN <=	pixel_green;
BLUE <=		pixel_blue;

-- PCLK-debounce & sync with SYSCLK50 for external PCLK
-- assuming PCLK is much slower than SYSCLK50 - not necessary if PCLK generated FPGA internally
process(SYSCLK50) begin
if rising_edge(SYSCLK50) then	
	pclk_intern <= PCLK_i;
end if;
end process;

-- assuming PCLK_i is generated internally
--pclk_intern <= PCLK_i;

-- ###############################################################################################################

sync_count: block	-- TFT-Timing, vom Display oder externer Quelle vorgegeben

component mem_color_tft
	port (
	clka: IN std_logic;
	dina: IN std_logic_VECTOR(3 downto 0);	-- R(2), G(1), B(0)
	addra: IN std_logic_VECTOR(10 downto 0);
	wea: IN std_logic_VECTOR(0 downto 0);
	douta: OUT std_logic_VECTOR(3 downto 0);	-- R(2), G(1), B(0)
	clkb: IN std_logic;
	dinb: IN std_logic_VECTOR(3 downto 0);
	addrb: IN std_logic_VECTOR(10 downto 0);
	web: IN std_logic_VECTOR(0 downto 0);
	doutb: OUT std_logic_VECTOR(3 downto 0));
end component;

component mem_tft
 port (
	clka: IN std_logic;
	dina: IN std_logic_VECTOR(7 downto 0);
	addra: IN std_logic_VECTOR(10 downto 0);
	wea: IN std_logic_VECTOR(0 downto 0);
	douta: OUT std_logic_VECTOR(7 downto 0);
	clkb: IN std_logic;
	dinb: IN std_logic_VECTOR(7 downto 0);
	addrb: IN std_logic_VECTOR(10 downto 0);
	web: IN std_logic_VECTOR(0 downto 0);
	doutb: OUT std_logic_VECTOR(7 downto 0));
end component;

component mem_font_tft
 port (
 clka: IN std_logic;
 addra: IN std_logic_VECTOR(11 downto 0);
 douta: OUT std_logic_VECTOR(7 downto 0));
end component;

signal v_porch,h_porch : STD_LOGIC_VECTOR(9 DOWNTO 0) :=(others =>'0'); -- Positionszhler
signal text_h, text_h_temp: STD_LOGIC_VECTOR(6 DOWNTO 0) :=(others =>'0');
signal text_v : STD_LOGIC_VECTOR(5 DOWNTO 0) :=(others =>'0');
signal font_line_count : STD_LOGIC_VECTOR(3 DOWNTO 0) :="0000";
signal font_pixel_select, font_pixel_count, font_pixel_temp: STD_LOGIC_VECTOR(3 DOWNTO 0) :="0000";
-- Text Buffer RAM Memory Signals, Port B (to VGA core)
signal text_data : std_logic_vector(07 downto 0);
signal text_adr_WR, text_adr : std_logic_vector(12 downto 0);
-- Font Buffer RAM Memory Signals
signal font_adr : std_logic_vector(11 downto 0);
signal font_data : std_logic_vector(07 downto 0);
-- Color Buffer RAM Memory Signals, Port B (to CPU core)
signal color_data, color_temp : std_logic_vector(03 downto 0);


type stateT is (state_wait_vstart,state_wait_hstart,state_count,state_scanend,state_stop);
signal state_timing : stateT := state_stop;
signal pixel_raw : std_logic;

-- Hardware Cursor
signal small   : std_logic;
signal curen2  : std_logic;
signal slowclk : std_logic;
signal curen0  : std_logic;
signal curen1    : std_logic;
signal blinkcounter : STD_LOGIC_VECTOR(3 downto 0);  
signal cur_en    : std_logic;
signal cur_mode  : std_logic;
signal cur_blink : std_logic;

begin

U_MEM: mem_tft port map (			-- Text RAM 2K x 8
	clka  => SYSCLK50,
	dina  => RAMDASCII,
	addra => text_adr_WR(10 downto 0),
	wea(0) => TEXTRAM_WR,
	douta => READASCII,
	clkb  => pclk_intern,
	dinb  => x"00",
	addrb => text_adr(10 downto 0),
	web(0) => '0',
	doutb => text_data
 );
U_COLOR: mem_color_tft port map (		-- COLOR RAM 2K x 4
	clka  => SYSCLK50,
	dina  => COLOR,	-- R(2), G(1), B(0)
	addra => text_adr_WR(10 downto 0),
	wea(0)   => TEXTRAM_WR,
	douta => READCOLOR,
	clkb  => pclk_intern,
	dinb  => "0000",
	addrb => text_adr(10 downto 0),
	web(0)   => '0',
	doutb => color_temp	-- R(2), G(1), B(0)
 );
U_FONT: mem_font_tft port map (
	clka => pclk_intern,
	addra => font_adr,
	douta => font_data); --open); -- 

text_adr_WR <=  "00" & TEXTY(4 downto 0) & TEXTX(5 downto 0);		-- 6 Bit Spalte (0..59 bzw. 63), 5 Bit Zeile (0..19 bzw. 31),
text_adr <= "00" & text_v(4 downto 0) & text_h_temp(5 downto 0);	-- Adressberechnung Text-RAM mit "Lchern"

font_adr <= ('0' & text_data & "000") + ("00" & text_data & "00") + font_line_count;	-- 12 Zeilen pro Zeichen, =text_data*8+text_data*4


-- Pixel Multiplexer, Pixel muss mit fallender PCLK-Flanke gltig sein
pixel_raw <= font_data(conv_integer(font_pixel_select(2 downto 0)));

-- Pixel Multiplexer, Pixel muss mit fallender PCLK-Flanke gltig sein
--pixel_raw <= font_data(0) WHEN font_pixel_select ="0000" ELSE 
--			font_data(1) WHEN font_pixel_select ="0001" ELSE
--			font_data(2) WHEN font_pixel_select ="0010" ELSE
--			font_data(3) WHEN font_pixel_select ="0011" ELSE
--			font_data(4) WHEN font_pixel_select ="0100" ELSE
--			font_data(5) WHEN font_pixel_select ="0101" ELSE
--			font_data(6) WHEN font_pixel_select ="0110" ELSE
--			font_data(7) WHEN font_pixel_select ="0111" ELSE
--			'0'; 
			
pixel <= (pixel_raw xor cursor_on xor color_data(3)) and frame_intern;
-- R(2), G(1), B(0)
pixel_red <= pixel and color_data(2);
pixel_green <= pixel and color_data(1);
pixel_blue <= pixel and color_data(0);


process(pclk_intern, RST, vstart_trigger) begin
if RST='1' or vstart_trigger='1' then -- Reset oder vstart_trigger-Impuls
	font_line_count<=(others =>'0');
	text_v<=(others =>'0');
	text_h <=(others =>'0');
	text_h_temp <=(others =>'0');
	v_porch<=(others =>'0');
	pixel_column<=(others =>'0');
	pixel_row<=(others =>'0');
	pixel_adr_intern <= (others =>'0');
	h_valid <= '0';
	v_valid <= '0';
	state_timing <= state_wait_vstart;
elsif rising_edge(pclk_intern) then
-- Statemachine, counter vert and hor
	color_data <= color_temp;
	font_pixel_select <= font_pixel_temp;		-- because of RAM latency delayed by two clocks
	font_pixel_temp <= font_pixel_count;
	text_h <= text_h_temp;						-- because of RAM latency delayed by two clocks
	frame_intern <= h_valid and v_valid;
	case state_timing is
		when state_wait_vstart =>				-- wait for vertical begin of frame
			if v_porch = ofs_v then
				v_valid <= '1';
				state_timing <= state_wait_hstart;
			end if;
			if hstart = '1' then
				v_porch <= v_porch +1;
			end if;			
		when state_wait_hstart =>				-- wait for horizontal begin of frame
			font_pixel_count <= (others =>'0');
			text_h_temp <=(others =>'0');
			pixel_column <= (others =>'0');
			if hstart = '1' then
				state_timing <= state_count;
			end if;
			if font_line_count = char_height then	-- check for font_line overflow
				text_v <= text_v +1;				-- yes, increment vertical text position
				font_line_count<=(others =>'0');
			end if;
		when state_count =>
			if pixel_column = screen_width  then
				h_valid <= '0';
				state_timing <= state_scanend;
			else
				h_valid <= '1';
				font_pixel_count <= font_pixel_count +1;
				if font_pixel_count = (char_width-1) then
					text_h_temp <= text_h_temp +1;	-- for cursor enable
					font_pixel_count <= (others =>'0');
				end if;
				pixel_column <= pixel_column +1;		
				pixel_adr_intern <= pixel_adr_intern + 1;
			end if;
		when state_scanend =>
			font_line_count <= font_line_count +1;	-- will be checked on hstart for overflow
			pixel_row <= pixel_row +1;
			if pixel_row = screen_height  then
				state_timing <= state_stop;			-- frame ended, wait for vstart_trigger pulse
			else
				state_timing <= state_wait_hstart;	-- wait for hstart pulse
			end if;
		when state_stop =>
			v_valid <= '0';
			pixel_column <= (others =>'0');
			pixel_row <= (others =>'0');
			pixel_adr_intern <= (others =>'0');
	end case;	

end if;
end process;

process(pclk_intern) begin
if falling_edge(pclk_intern) then	
-- VSTART_i und HSTART_i change on rising edge on digital picture frame
-- one-shots for vstart_trigger und HSTART
	vstart_trigger<='0';
	if (vstart_d='0') and (VSTART_i='1') then
		vstart_trigger<='1';
		vstart_preset<='1';
	end if;
	vstart_d<=VSTART_i;
	h_porch<=h_porch+1;
	if (hstart_d='0') and (HSTART_i='1') then
		h_porch<=(others =>'0');
	end if;
	hstart_d<=HSTART_i;
	hstart<='0';
	vstart<='0';
	if h_porch=ofs_h then
		hstart<='1';	-- trigger hstart if h_porch position reached
		if vstart_preset='1' and state_timing = state_wait_hstart then
			vstart<='1';
			vstart_preset<='0';
		end if;
	end if;
end if;
end process;


-- ###############################################################################################################

-- Hardware Cursor

	cur_mode  <= CURSTYLE(0);
	cur_blink <= CURSTYLE(1);
	cur_en    <= CURSTYLE(2);
	-- slowclk for blink hardware cursor
	blinkcounter <= blinkcounter + 1 when rising_edge(vstart_trigger);
	slowclk <= blinkcounter(3); --5 Hz

	curen0 <= '1' when (text_v = TEXTY) and (text_h = TEXTX) else '0';
	small  <= '1' when (font_line_count > 8) else '0';
	curen2 <= (slowclk or (not cur_blink)) and cur_en;
	curen1 <= '1' when cur_mode = '0' else small;
	cursor_on <= frame_intern and curen0 and curen1 and curen2;

    
end block;
-- ###############################################################################################################
  
end Behavioral;
